

from typing import Optional, Dict
# import datetime
import os
import traceback
import time
import tarfile

from sqlalchemy.orm import sessionmaker
import pandas as pd
import numpy as np

from ...util.mail_retriever import MailAttachmentRetriever, IMAP_SPType, UID_FILE_NAME
from ...util.wechat_bot import WechatBot
from ...util.calculator import Calculator
from ..wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from ..api.raw import RawDataApi
from ..api.basic import BasicDataApi
from ..api.derived import DerivedDataApi
from ..view.raw_models import HFIndexPrice, HFFundNav
from ..view.basic_models import FOFInfo
from ..view.derived_models import FOFNavPublic
from ..fund.raw.other_raw_data_downloader import OtherRawDataDownloader
from ..fund.raw.raw_data_helper import RawDataHelper


class ResearchNAVReader:

    def __init__(self, read_dir: str, user_name: str, password: str):
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)

        self._read_dir = read_dir
        assert os.path.isdir(self._read_dir), f'arg dump_dir should be a directory (now){self._read_dir}'

        self._user_name = user_name
        self._password = password
        self._wechat_bot = WechatBot()

    @staticmethod
    def _read_for_NAVData(file_path: str) -> Dict[str, int]:
        update_info: Dict[str, int] = {}
        tar = tarfile.open(file_path, 'r:gz', encoding='gb2312')
        for one in tar:
            if not one.name.endswith('.xls') and not one.name.endswith('.xlsx'):
                continue
            if '业绩走势' in one.name:
                continue
            try:
                try:
                    normal_replace_dict = {
                        '・': '·',
                    }
                    p_name = one.name
                    for k, v in normal_replace_dict.items():
                        p_name = p_name.replace(k, v)
                    print(f'[_read_for_NAVData] (name in tar){one.name}')
                except UnicodeEncodeError:
                    replace_dict = {
                        '\udc9bK': '汯',
                        '\udcacB': '珺',
                        '\udcd4Z': '訸',
                        '\udcb6G': '禛',
                        '\udc95D': '旸',
                        '\udcf1I': '馡',
                        '\udca9q': '﹒',
                    }
                    p_name = one.name
                    for k, v in replace_dict.items():
                        p_name = p_name.replace(k, v)
                    print(f'[_read_for_NAVData] (name in tar (fixed)){p_name}')
                p_name = p_name.split('.')[0]
                df = pd.read_excel(tar.extractfile(one), na_values=['--'])
                df = df.rename(columns={'日期': 'datetime', '净值(分红再投)': 'nav'})
                df = df[['datetime', 'nav']]
                hf_fund_info = RawDataApi().get_hf_fund_info()
                hf_fund_info = hf_fund_info[hf_fund_info.desc_name == p_name]
                fund_id = hf_fund_info.fund_id.array[0]
                df['fund_id'] = fund_id
                df['datetime'] = pd.to_datetime(df.datetime, infer_datetime_format=True).dt.date
                hf_fund_nav = RawDataApi().get_hf_fund_nav(fund_ids=[fund_id])
                if hf_fund_nav is not None:
                    df_fitted = df.astype(hf_fund_nav.dtypes.to_dict())
                    # merge on all columns
                    df_fitted = df_fitted.merge(hf_fund_nav, how='left', indicator=True, validate='one_to_one')
                    df_fitted = df_fitted[df_fitted._merge == 'left_only'].drop(columns=['_merge'])
                else:
                    df_fitted = df
                print(f'to hf_fund_nav {df_fitted}')
                # TODO: 最好原子提交下边两步
                RawDataApi().delete_hf_fund_nav(fund_id_to_delete=fund_id, datetime_list=df_fitted.datetime.to_list())
                # 更新到db
                df_fitted.to_sql(HFFundNav.__table__.name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')

                # 再向 FOFNavPublic 更新一份
                df = df.rename(columns={'fund_id': 'fof_id', 'nav': 'adjusted_nav'})
                df['nav'] = np.nan
                df['acc_net_value'] = np.nan
                fof_nav_public = DerivedDataApi().get_fof_nav_public(fof_id_list=[fund_id])
                if fof_nav_public is not None:
                    fof_nav_public = fof_nav_public.drop(columns=['volume', 'mv', 'ret', 'create_time', 'update_time', 'is_deleted'])
                    df = df.astype(fof_nav_public.dtypes.to_dict())
                    # merge on all columns
                    df = df.merge(fof_nav_public, how='left', indicator=True, validate='one_to_one')
                    df = df[df._merge == 'left_only'].drop(columns=['_merge'])
                print(f'to fof_nav_public {df}')
                # TODO: 最好原子提交下边两步
                DerivedDataApi().delete_fof_nav_public(fof_id_to_delete=fund_id, datetime_list=df.datetime.to_list())
                # 更新到db
                df.to_sql(FOFNavPublic.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
                update_info[p_name] = df.shape[0]

                fof_nav_public = DerivedDataApi().get_fof_nav_public(fof_id_list=[fund_id])
                if fof_nav_public is not None:
                    fof_nav_public = fof_nav_public.set_index('datetime')['adjusted_nav']
                    res_status = Calculator.get_stat_result(fof_nav_public.index, fof_nav_public.array)
                    Session = sessionmaker(BasicDatabaseConnector().get_engine())
                    db_session = Session()
                    # '1' 表示公有的 FOFInfo
                    fof_info_to_set = db_session.query(FOFInfo).filter(FOFInfo.manager_id == '1', FOFInfo.fof_id == fund_id).one_or_none()
                    fof_info_to_set.net_asset_value = float(res_status.last_unit_nav) if not pd.isnull(res_status.last_unit_nav) else None
                    fof_info_to_set.adjusted_net_value = float(fof_nav_public.array[-1]) if not pd.isnull(fof_nav_public.arraay[-1]) else None
                    fof_info_to_set.latest_cal_date = res_status.end_date
                    fof_info_to_set.ret_year_to_now = float(res_status.recent_year_ret) if not pd.isnull(res_status.recent_year_ret) else None
                    fof_info_to_set.ret_total = float(res_status.cumu_ret) if not pd.isnull(res_status.cumu_ret) else None
                    fof_info_to_set.ret_ann = float(res_status.annualized_ret) if not pd.isnull(res_status.annualized_ret) else None
                    fof_info_to_set.mdd = float(res_status.mdd) if not pd.isnull(res_status.mdd) else None
                    fof_info_to_set.sharpe = float(res_status.sharpe) if not pd.isnull(res_status.sharpe) else None
                    fof_info_to_set.vol = float(res_status.annualized_vol) if not pd.isnull(res_status.annualized_vol) else None
                    db_session.commit()
                    db_session.close()
            except Exception as e:
                traceback.print_exc()
                print(f'[_read_for_NAVData] failed (err){e}')
        tar.close()
        return update_info

    @staticmethod
    def _read_for_IndexData(file_path: str) -> Dict[str, int]:
        update_info = {}
        tar = tarfile.open(file_path, 'r:gz', encoding='gb2312')
        for one in tar:
            print(f'[_read_for_IndexData] (name in tar){one.name}')
            index_name = one.name.split('/')[-1].split('.')[0]
            if '朝阳永续' in one.name:
                key_name = one.name.split('朝阳永续-')[-1].split('.')[0]
                df = pd.read_excel(tar.extractfile(one), na_values=['--'], thousands=',')
                df = df.rename(columns={'时间': 'index_date', key_name: 'close'})
                df = df[['index_date', 'close']]
                df['calcu_date'] = None
            else:
                if '融智' not in one.name:
                    continue
                    # print(f'[_read_for_IndexData] got special file name {one}')
                key_name = one.name.split('融智-')[-1].split('.')[0]
                df = pd.read_excel(tar.extractfile(one), header=3, na_values=['--'])
                df = df.rename(columns={'指数日期': 'index_date', '指数计算日期': 'calcu_date', key_name: 'close'})
                df = df[['index_date', 'calcu_date', 'close']]
                df['calcu_date'] = pd.to_datetime(df.calcu_date, infer_datetime_format=True).dt.date
            df = df[df.close.notna()]
            asset_info = BasicDataApi().get_asset_info_by_type(['策略指数'])
            asset_info = asset_info[asset_info.real_name == index_name]
            if asset_info.empty:
                assert False, f'[_read_for_IndexData] can not find name of index {index_name}'
            index_id = asset_info.real_id.array[0]
            df['index_id'] = index_id
            df['index_date'] = pd.to_datetime(df.index_date, infer_datetime_format=True).dt.date
            hf_index_price = RawDataApi().get_hf_index_price(index_ids=[index_id])
            if hf_index_price is not None:
                df = df.astype(hf_index_price.dtypes.to_dict())
                # merge on all columns
                df = df.merge(hf_index_price, how='left', indicator=True, validate='one_to_one')
                df = df[df._merge == 'left_only'].drop(columns=['_merge'])
            print(df)
            # TODO: 最好原子提交下边两步
            RawDataApi().delete_hf_index_price(index_id, df.index_date.to_list())
            # 更新到db
            df.to_sql(HFIndexPrice.__table__.name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')
            update_info[key_name] = df.shape[0]
        tar.close()
        return update_info

    def _notify_error_event(self, err_msg: str):
        print(f'[read_navs_and_dump_to_db] {err_msg}')
        # self._wechat_bot.send_hedge_fund_nav_update_failed(err_msg)

    def read_navs_and_dump_to_db(self):
        try:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'rb') as f:
                uid_last = f.read()
                if not uid_last:
                    uid_last = None
        except Exception as e:
            self._notify_error_event(f'read uid file failed (e){e}, use None instead(read all emails)')
            uid_last = None

        try:
            mar = MailAttachmentRetriever(self._read_dir, ['tar.gz', ])
            data = mar.get_excels(IMAP_SPType.IMAP_QQ, self._user_name, self._password, uid_last)
        except Exception as e:
            self._notify_error_event(f'FATAL ERROR!! get new data of hedge fund nav failed (e){e}')
            return

        uid_last_succeed: Optional[bytes] = None
        for name, comp_date in data.items():
            uid, file_path = comp_date
            if not name.endswith('.tar.gz'):
                continue

            try:
                if '净值数据' in name:
                    update_info = ResearchNAVReader._read_for_NAVData(file_path)

                    ordd = OtherRawDataDownloader(RawDataHelper())
                    ret = ordd.hf_fund_nav_miss()
                    assert not ret, 'update table of hf fund nav miss failed!!'

                    self._wechat_bot.send_research_data_update('净值', update_info)
                elif '指数数据' in name:
                    update_info = ResearchNAVReader._read_for_IndexData(file_path)

                    self._wechat_bot.send_research_data_update('指数', update_info)
                else:
                    raise NotImplementedError('unknown research nav file from attachment')
            except Exception as e:
                traceback.print_exc()
                self._notify_error_event(f'{e} (parse) (name){name} (file_path){file_path}')
                continue

            uid_last_succeed = uid
            time.sleep(1)

        # 记录下成功的最后一个uid
        if uid_last_succeed is not None:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'wb') as f:
                f.write(uid_last_succeed)
        return


if __name__ == '__main__':
    try:
        email_data_dir = os.environ['SURFING_EMAIL_DATA_DIR']
        user_name = os.environ['SURFING_EMAIL_USER_NAME']
        password = os.environ['SURFING_EMAIL_PASSWORD']
    except KeyError as e:
        import sys
        sys.exit(f'can not found enough params in env (e){e}')

    r_nav_r = ResearchNAVReader(email_data_dir, user_name, password)
    r_nav_r.read_navs_and_dump_to_db()




from typing import Dict

import pandas as pd
import numpy as np

from ...data.manager import DataManager
from .factor_types import VolatilityFactor, MixFactorBase
from .basic_factors import StockPostPriceFactor
from .utils import _TRADING_DAYS_PER_YEAR, normalize


class BetaSSEIFactor(VolatilityFactor):

    _WINDOW = 30

    # 上证指数beta
    def __init__(self):
        super().__init__('b_ssei')
        self._deps.add(StockPostPriceFactor())

    def calc(self):
        def _rolling_regression(x: np.ndarray):
            data = combined.iloc[x, :]
            self._factor.iloc[int(x[-1]), :] = np.polyfit(x=data.sse, y=data.drop(columns='sse'), deg=1)[0]
            return np.nan

        # 股票对上证指数回归得到的贝塔
        stock_ret = StockPostPriceFactor().get().pct_change()
        index_price = DataManager.basic_data('get_index_price_dt', index_list=['sse', 'tmd_1y'], columns=['close'])
        index_price = index_price.pivot(index='datetime', columns='index_id', values='close')
        index_price = index_price.reindex(index_price.index.union(stock_ret.index)).ffill().reindex(stock_ret.index)

        # 日化一年整存整取利率
        tmd_1y = index_price.tmd_1y * 0.01 / _TRADING_DAYS_PER_YEAR
        pd.testing.assert_index_equal(stock_ret.index, index_price.index)

        combined = stock_ret.join(index_price.sse.pct_change())
        combined = combined.sub(tmd_1y, axis=0).iloc[1:, :]

        self._factor = pd.DataFrame(index=combined.index, columns=stock_ret.columns)
        pd.Series(range(combined.shape[0])).rolling(window=self._WINDOW).apply(_rolling_regression, raw=True)
        self._factor = self._factor.reindex(stock_ret.index)


class Vol250DFactor(VolatilityFactor):

    _WINDOW = 250

    # 波动率（250日）
    def __init__(self):
        super().__init__('v_250d')
        self._deps.add(StockPostPriceFactor())

    def calc(self):
        # 股票收益率的标准差，选取过去250日的收益率进行计算
        stock_ret = StockPostPriceFactor().get().pct_change()
        self._factor = stock_ret.rolling(window=self._WINDOW).std()


class ResidualVolSSEIFactor(VolatilityFactor):

    _WINDOW = 30

    # 上证指数残差波动率
    def __init__(self):
        super().__init__('rv_ssei')
        self._deps.add(StockPostPriceFactor())

    def calc(self):
        def _rolling_regression(x: np.ndarray):
            data = combined.iloc[x, :]
            # 计算残差的标准差
            self._factor.iloc[int(x[-1]), :] = np.sqrt(np.polyfit(x=data.sse, y=data.drop(columns='sse'), deg=1, full=True)[1] / (self._WINDOW - 1))
            return np.nan

        # 股票对上证指数回归得到的贝塔
        stock_ret = StockPostPriceFactor().get().pct_change()
        index_price = DataManager.basic_data('get_index_price_dt', index_list=['sse', 'tmd_1y'], columns=['close'])
        index_price = index_price.pivot(index='datetime', columns='index_id', values='close')
        index_price = index_price.reindex(index_price.index.union(stock_ret.index)).ffill().reindex(stock_ret.index)

        # 日化一年整存整取利率
        tmd_1y = index_price.tmd_1y * 0.01 / _TRADING_DAYS_PER_YEAR
        pd.testing.assert_index_equal(stock_ret.index, index_price.index)

        combined = stock_ret.join(index_price.sse.pct_change())
        combined = combined.sub(tmd_1y, axis=0).iloc[1:, :]

        self._factor = pd.DataFrame(index=combined.index, columns=stock_ret.columns)
        pd.Series(range(combined.shape[0])).rolling(window=self._WINDOW).apply(_rolling_regression, raw=True)
        self._factor = self._factor.reindex(stock_ret.index)


class MixVolatilityFactor(MixFactorBase, VolatilityFactor):
    # 合成因子
    def __init__(self):
        super(MixFactorBase, self).__init__('mixv')
        self._factor: Dict[pd.DataFrame] = {}
        self._deps.add(BetaSSEIFactor())
        self._deps.add(Vol250DFactor())
        self._deps.add(ResidualVolSSEIFactor())

    def calc(self, universe: str):
        index = BetaSSEIFactor().get(universe).index
        columns = BetaSSEIFactor().get(universe).columns
        self._factor[universe] = pd.DataFrame((normalize(BetaSSEIFactor().get(universe).T) + normalize(Vol250DFactor().get(universe).T) +
                                               normalize(ResidualVolSSEIFactor().get(universe).T)).T, index=index, columns=columns)



import pandas as pd
import numpy as np
from ....manager.data_tables import StockInfoDataTables
from .constants import ROE有效阈值


def cut_out_roe(data):
    if np.isnan(data):
        return np.nan

    if data < -ROE有效阈值:
        return -ROE有效阈值
    if data > ROE有效阈值:
        return ROE有效阈值

    return data


def average(key, quarter_shift=1):
    # 计算指标跟上一期的平均值 todo 这里因为财报不足4期，先用1期数据
    return (key + key.shift(quarter_shift).fillna(method='pad'))/2


def unify_all_data(sdt):
    # todo 借壳上市处理
    index = sdt.income_ttm.index

    income_ttm = sdt.income_ttm
    income_q = sdt.income_q.reindex(index)[[
        '单季度.营业总收入',
        '单季度.营业利润',
        '单季度.利润总额',
        '单季度.归属于母公司股东的净利润',
    ]]

    cash_flow_ttm = sdt.cash_flow_ttm.reindex(index)
    cash_flow_q = sdt.cash_flow_q.reindex(index)[[
        '单季度.经营活动产生的现金流量净额',
        '单季度.购建固定资产、无形资产和其他长期资产所支付的现金',
    ]]
    balance = sdt.balance.reindex(index)

    df = pd.concat([income_ttm, cash_flow_ttm, balance, income_q, cash_flow_q], axis=1)
    df['期初净资产'] = df['归属于母公司股东权益合计'].shift(1)
    return df


def compute_ttm_quarterly_ratios(sdt: StockInfoDataTables):
    df = unify_all_data(sdt)
    df = df.sort_index()
    df = df.fillna(0)

    # 盈利能力指标计算
    df['净资产'] = df['归属于母公司股东权益合计']
    df['TTM.期初ROE'] = df['TTM.归属于母公司股东的净利润'] / df['期初净资产']
    df['TTM.期末ROE'] = df['TTM.归属于母公司股东的净利润'] / df['净资产']
    df['平均净资产'] = (df['期初净资产'] + df['净资产']) / 2
    df['TTM.平均ROE'] = 2 * df['TTM.归属于母公司股东的净利润'] / (df['期初净资产'] + df['净资产'])

    df['TTM.期初ROE'] = df['TTM.期初ROE'].apply(cut_out_roe)
    df['TTM.期末ROE'] = df['TTM.期末ROE'].apply(cut_out_roe)
    df['TTM.平均ROE'] = df['TTM.平均ROE'].apply(cut_out_roe)

    df['平均总资产'] = average(df['资产总计'])
    df['TTM.ROA'] = df['TTM.归属于母公司股东的净利润'] / df['平均总资产']

    df['NOPAT'] = (df['TTM.营业利润'] + df['TTM.财务费用']) * (1 - df['TTM.所得税费用'] / df['TTM.利润总额'])
    df['无息流动负债'] = df['应付票据'] \
                   + df['应付账款'] \
                   + df['预收款项'] \
                   + df['应付职工薪酬'] \
                   + df['应交税费'] \
                   + df['其他应付款合计'] \
                   + df['一年内的递延收益'] \
                   + df['其他流动负债']

    df['无息非流动负债'] = df['长期应付款'] \
                    + df['专项应付款'] \
                    + df['预计非流动负债'] \
                    + df['递延收益'] \
                    + df['递延所得税负债'] \
                    + df['其他非流动负债']
    df['投资资本'] = df['资产总计'] - df['无息流动负债'] - df['商誉']
    df['税后净营业利润'] = df['NOPAT']
    df['TTM.ROIC'] = df['NOPAT'] / df['投资资本']

    df['TTM.计算毛利润营业收入'] = \
        df['TTM.营业收入'] \
        + df['TTM.利息收入'] \
        + df['TTM.已赚保费'] \
        + df['TTM.手续费及佣金收入'] \
        + df['TTM.营业总收入其他项目']
    df['TTM.计算营业成本'] = df['TTM.营业成本'] + \
                   df['TTM.利息支出'] + \
                   df['TTM.手续费及佣金支出'] + \
                   df['TTM.研发费用'] + \
                   df['TTM.退保金'] + \
                   df['TTM.赔付支出净额'] + \
                   df['TTM.提取保险合同准备金净额'] + \
                   df['TTM.保单红利支出'] + \
                   df['TTM.分保费用'] + \
                   df['TTM.其他业务成本']
    df['TTM.毛利润'] = df['TTM.计算毛利润营业收入'] - df['TTM.计算营业成本']
    df['TTM.毛利润率'] = df['TTM.毛利润'] / df['TTM.营业总收入']
    df['TTM.净利润率'] = df['TTM.归属于母公司股东的净利润'] / df['TTM.营业总收入']
    df['TTM.研发费用率'] = df['TTM.研发费用'] / df['TTM.营业总收入']

    # 现金流量
    df['TTM.自由现金流'] = df['TTM.经营活动产生的现金流量净额'] - df['TTM.购建固定资产、无形资产和其他长期资产所支付的现金']
    df['TTM.自由现金流对营业收入比'] = df['TTM.自由现金流'] / df['TTM.营业总收入']
    df['TTM.自由现金流对净利润比'] = df['TTM.自由现金流'] / df['TTM.归属于母公司股东的净利润']
    df['TTM.资本支出'] = df['TTM.投资所支付的现金'] + df['TTM.购建固定资产、无形资产和其他长期资产所支付的现金'] + \
                 df['TTM.取得子公司及其他营业单位支付的现金净额'] + df['TTM.支付的其他与投资活动有关的现金']
    df['TTM.资本支出比营业收入'] = df['TTM.资本支出'] / df['TTM.营业总收入']
    df['TTM.资本支出比净利润'] = df['TTM.资本支出'] / df['TTM.归属于母公司股东的净利润']
    df['TTM.经营现金流比资本支出'] = df['TTM.经营活动产生的现金流量净额'] / df['TTM.资本支出']
    # df['TTM.折旧摊销'] = df['TTM.固定资产和投资性房地产折旧'] + df['TTM.无形资产摊销'] + df['TTM.长期待摊费用摊销']
    # df['TTM.折旧摊销比经营现金流'] = df['TTM.折旧摊销'] / df['TTM.经营活动产生的现金流量净额']
    # df['TTM.资本支出比折旧摊销'] = df['TTM.资本支出'] / df['TTM.折旧摊销']

    # 收益质量
    df['TTM.经营现金流净值比利润总额'] = df['TTM.经营活动产生的现金流量净额'] / df['TTM.利润总额']
    df['TTM.所得税比利润总额'] = df['TTM.所得税费用'] / df['TTM.利润总额']
    df['TTM.营业外支出比利润总额'] = df['TTM.营业外支出'] / df['TTM.利润总额']

    # 运营效率
    df['总资产周转天数'] = (df['平均总资产'] / df['TTM.营业总收入']) * 365
    df['总资产周转次数'] = df['TTM.营业总收入'] / df['平均总资产']

    df['应收'] = df['应收票据'] + \
               df['应收账款'] + \
               df['应收保费'] + \
               df['应收分保账款'] + \
               df['应收分保合同准备金'] + \
               df['其他应收款合计'] + \
               df['应收出口退税'] + \
               df['应收补贴款'] + \
               df['内部应收款'] + \
               df['长期应收款'] - \
               df['预付款项']

    df['年均应收'] = average(df['应收'])
    df['应收账款周转次数'] = df['TTM.营业总收入'] / df['年均应收']
    df['应收账款周转天数'] = (df['年均应收'] / df['TTM.营业总收入']) * 365

    df['应付'] = df['应付票据'] + \
                df['应付账款'] + \
                df['应付手续费及佣金'] + \
                df['应付职工薪酬'] + \
                df['应交税费'] + \
                df['其他应付款合计'] + \
                df['内部应付款'] + \
                df['保险合同准备金'] + \
                df['长期应付款'] + \
                df['专项应付款'] + \
                df['长期应付职工薪酬'] - \
                df['预付款项']
    df['年均应付'] = average(df['应付'])
    df['应付账款周转次数'] = df['TTM.营业成本'] / df['年均应付']
    df['应付账款周转天数'] = (df['年均应付'] / df['TTM.营业成本']) * 365
    df['年均存货'] = average(df['存货'])
    df['存货周转次数'] = df['TTM.营业成本'] / df['年均存货']
    df['存货周转天数'] = (df['年均存货'] / df['TTM.营业成本']) * 365
    df['运营周转天数'] = df['应收账款周转天数'] + df['存货周转天数']
    df['现金周转天数'] = df['应收账款周转天数'] + \
                   df['存货周转天数'] - \
                   df['应付账款周转天数']

    # 资本结构
    df['流动比率'] = df['流动资产合计'] / df['流动负债合计']
    df['利息保障倍数'] = (df['TTM.利润总额'] + df['TTM.财务费用']) / df['TTM.财务费用']
    df['流动资产比总资产'] = average(df['流动资产合计']) / df['平均总资产']
    df['非流动资产比总资产'] = average(df['非流动资产合计']) / df['平均总资产']
    df['净营运资本'] = df['应收票据'] + df['应收账款'] + df['预付款项'] + df['其他应收款合计'] + df['存货'] - df['无息流动负债'] + df['长期股权投资'] + df[
        '投资性房地产']
    df['有形资本'] = df['净营运资本'] + df['固定资产'] + df['在建工程']
    df['投资资本'] = df['资产总计'] - df['无息流动负债'] - df['商誉']
    df['有型资产比总资产'] = average(df['有形资本']) / df['平均总资产']
    df['资产负债率'] = average(df['负债合计']) / df['平均总资产']
    df['有息负债'] = df['负债合计'] - df['无息流动负债'] - df['无息非流动负债']
    df['有息负债率'] = average(df['有息负债']) / df['平均总资产']

    # 成长能力
    df['单季度.自由现金流'] = df['单季度.经营活动产生的现金流量净额'] - df['单季度.购建固定资产、无形资产和其他长期资产所支付的现金']
    df['单季度.自由现金流年成长率'] = df['单季度.自由现金流'].pct_change(periods=4)
    df['近1年净利润增速'] = df['TTM.归属于母公司股东的净利润'].pct_change(periods=4)

    return df[[
        'TTM.期初ROE',
        'TTM.期末ROE',
        'TTM.平均ROE',
        'TTM.ROA',
        'TTM.ROIC',
        'TTM.毛利润率',
        'TTM.研发费用率',

        'TTM.自由现金流对净利润比',
        'TTM.资本支出比净利润',
        # 'TTM.折旧摊销比经营现金流',
        # 'TTM.资本支出比折旧摊销',

        'TTM.经营现金流净值比利润总额',
        'TTM.所得税比利润总额',
        'TTM.营业外支出比利润总额',

        '总资产周转天数',
        '总资产周转次数',
        '应收账款周转天数',
        '应收账款周转次数',
        '应付账款周转天数',
        '应付账款周转次数',
        '存货周转天数',
        '存货周转次数',
        '现金周转天数',

        '流动比率',
        '利息保障倍数',
        '流动资产比总资产',
        '有型资产比总资产',
        '资产负债率',
        '有息负债率',
        '单季度.自由现金流年成长率',
        '近1年净利润增速',

        # prices ratios 需要的列
        '平均净资产',
        '平均总资产',
        '税后净营业利润',
        '投资资本',
        'TTM.归属于母公司股东的净利润',
        '净资产',
        'TTM.经营活动产生的现金流量净额',
        'TTM.营业总收入',
        'TTM.毛利润',
    ]]
